/****
 *
 * Ttulo: Calculando Raiz Quadrada usando o mtodo
 *         de Newton e Raphson
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2012
 * ltima modificao: 16/06/2012
 *
 * Descrio: Calcula a raz quadrada de um nmero
 *            real no negativo usando o mtodo de
 *            Newton e Raphson
 *
 * Entrada: Um nmero real
 *
 * Sada: A raiz quadrada do nmero
 *
 ****/

#include <stdio.h>   /* printf() */
#include <math.h>    /* sqrt()   */

   /* Tolerncia de erro admitida no clculo da raiz */
#define DELTA 1.0E-10

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
static int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeReal(): L um valor do tipo double via teclado
 *           deixando o buffer associado ao teclado
 *           sem caracteres remanescentes
 *
 * Parmetros: Nenhum
 *
 * Retorno: O valor lido e validado
 *
 ****/
double LeReal(void)
{
   int    teste, /* Armazena o retorno de scanf() */
          nResto = 0; /* Nmero de caracteres */
                      /* excedentes           */
   double valorLido; /* O valor digitado pelo usurio */

      /* O lao encerra quando o usurio se comporta bem */
   while (1) {
         /* Tenta ler um valor do tipo double */
      teste = scanf("%lf", &valorLido);

         /* Se no ocorreu erro de leitura ou de final */
         /* de arquivo, h caracteres remanescentes    */
         /* que precisam ser removidos                 */
      if (teste != EOF) {
         nResto = LimpaBuffer();
      }

         /* Se o valor retornado por scanf() */
         /* foi 1, a leitura foi OK          */
      if (teste == 1) {
         break; /* Leitura foi nota 10 */
      } else { /* Usurio foi mal comportado */
         printf( "\a\n\t>>> O valor digitado e' invalido. "
                 "Tente novamente\n\t> " );
      }
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf("\t>>> %d caracteres foram descartados\n",
             nResto);
   }

      /* Retorna um valor vlido do tipo double */
   return valorLido;
}

/****
 * ComparaDoubles(): Compara dois valores do tipo double
 *
 * Parmetros:
 *      d1, d2 (entrada): valores que sero comparados
 *
 * Retorno: 0, se os nmeros forem considerados iguais
 *          -1, se d1 for considerado menor do que d2
 *          1, se d1 for considerado maior do que d2
 ****/
int ComparaDoubles(double d1, double d2)
{
      /* Verifica se o valor absoluto da diferena entre  */
      /* os nmeros  menor do que ou igual  preciso    */
      /* determinada pela constante DELTA. Se for o caso, */
      /* eles so considerados iguais. Caso contrrio,    */
      /* verifica qual deles  menor ou maior.            */
   if (fabs(d1 - d2) <= DELTA) {
      return 0; /* d1 e d2 so considerados iguais */
   } else if (d1 < d2) {
      return -1; /* d1  menor do que d2 */
   } else {
      return 1; /* d1  maior do que d2 */
   }
}

/****
 *
 * SqrtNewton(): Calcula raiz quadrada usando
 *               o mtodo de Newton e Raphson
 *
 * Parmetros:
 *      x (entrada) - nmero real cuja raiz
 *                    quadrada ser calculada
 *
 * Retorno: A raiz quadrada do parmetro, se ele
 *          no for negativo. Caso contrrio, -1.0.
 *
 ****/
double SqrtNewton(double x)
{
   double raiz;
      /* A raiz quadrada de zero  zero */
   if (x == 0.0) {
      return 0.0;
   }
      /* Se o parmetro for negativo apresenta uma  */
      /* mensagem correspondente e retorna um valor */
      /* indicando o erro na chamada da funo      */
   if (x < 0) {
      printf("\nChamada incorreta da funcao\n");
      return -1.0;
   }
      /* A estimativa inicial da raiz  o prprio */
      /* valor cuja raiz se deseja calcular       */
   raiz = x;
      /* Enquanto o quadrado da estimativa no */
      /* for suficientemente prximo de x,     */
      /* continua calculando estimativas       */
   while (ComparaDoubles(x, raiz*raiz)) {
         /* Faz nova estimativa */
      raiz = (raiz + x/raiz)/2;
   }
   return raiz;
}

/****
 *
 * main(): Testa SqrtNewton() comparando o resultado apresentado
 *         por essa funo com aquele apresentado por sqrt()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   double num;

      /* Apresenta o programa */
   printf("\n\t>>> Este programa calcula a raiz quadrada "
          "\n\t>>> de um numero nao negativo usando "
          "\n\t>>> o metodo de Newton e Raphson.\n");

      /* L o nmero cuja raiz quadrada ser calculada */
   printf("\n\t>>> Digite um numero nao negativo> ");
   num = LeReal();

      /* Se a raiz quadrada existir, calcula-a e exibe-a */
   if(num < 0) { /* Usurio desobediente */
      printf("\n\t>>> Nao ha raiz quadrada de numero negativo\n");
   } else { /* Usurio bacana */
      printf( "\n\t>>> Raiz quadrada aproximada de %3.2f: "
              "%3.2f\n", num, SqrtNewton(num) );

         /* Apresenta o resultado calculado usando sqrt()*/
      printf( "\t>>> Raiz quadrada de %3.2f "
              "usando sqrt(): %3.2f\n", num, sqrt(num) );
   }

   return 0;
}

/******************* Exemplo de Execuo *******************

        >>> Este programa calcula a raiz quadrada
        >>> de um numero nao negativo usando
        >>> o metodo de Newton e Raphson.

        >>> Digite um numero nao negativo> 2

        >>> Raiz quadrada aproximada de 2.00: 1.41
        >>> Raiz quadrada de 2.00 usando sqrt(): 1.41

***********************************************************/
