/****
 *
 * Ttulo: Resolvendo Sudoku
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/12/2015
 * ltima modificao: 21/12/2015
 *
 * Descrio: Resolve um quebra-cabea Sudoku
 *
 * Entrada: No h
 *
 * Sada: O quebra-cabea resolvido
 *
 ****/

#include <stdio.h> /* printf() */

/****
 *
 * EstaDisponivel(): Verifica se um nmero pode ser usado no
 *                   quadrado especificado por linha e coluna
 *                   de uma grade de quebra-cabea Sudoku
 *
 * Parmetros:
 *      grade[][] (entrada) - array bidimensional que representa
 *                            o estado atual do quebra-cabea
 *      linha, coluna (entrada) - a linha e coluna que definem
 *                                o quadrado
 *      numero (entrada) - o nmero em questo
 *
 * Retorno: 1, se o nmero estiver disponvel para uso;
 *          0, em caso contrrio.
 *
 ****/
int EstaDisponivel( int grade[][9],
                    int linha, int coluna, int numero )
{
         /* Linha inicial de uma subgrade (0, 3 ou 6) */
   int linhaInicial = (linha/3) * 3,
         /* Coluna inicial de uma subgrade (0, 3 ou 6) */
       colunaInicial = (coluna/3) * 3;

         /***********************************************/
         /* Todo que quebra-cabea Sudoku usa uma grade */
         /* 9x9. Logo, 9 no  nenhum nmero mgico.    */
         /***********************************************/

   for(int i = 0; i < 9; ++i) {
         /* Verifica se o nmero j ocorre na   */
         /* linha em que o quadrado se encontra */
      if (grade[linha][i] == numero)
         return 0; /* O nmero j aparece na linha */

         /* Verifica se o nmero j ocorre na    */
         /* coluna em que o quadrado se encontra */
      if (grade[i][coluna] == numero)
         return 0; /* O nmero j aparece na coluna */

         /* Verifica se o nmero j ocorre na      */
         /* subgrade em que o quadrado se encontra */
      if (grade[linhaInicial + (i%3)][colunaInicial + (i/3)] ==
             numero)
         return 0; /* O nmero j aparece na subgrade */
   }

      /* O nmero est disponvel */
   return 1;
}

/****
 *
 * PreencheSudoku(): Preenche recursivamente cada quadrado
 *                   de uma grade que representa o
 *                   quebra-cabea Sudoku
 *
 * Parmetros:
 *      grade[][] (entrada e sada) - array que representa
 *                                    o estado atual do
 *                                    quebra-cabea
 *      linha, coluna (entrada) - linha e coluna que identificam
 *                                um quadrado a ser preenchido
 *
 * Retorno: 1, se a grade foi preenchida; 0, em caso contrrio.
 *
 ****/
int PreencheSudoku(int grade[][9], int linha, int coluna)
{
   int numero;

         /***********************************************/
         /* Todo que quebra-cabea Sudoku usa uma grade */
         /* 9x9. Logo, 9 no  nenhum nmero mgico.    */
         /***********************************************/

      /* Verifica se a linha e a coluna recebidas como  */
      /* parmetros fazem parte da grade. Se no for o  */
      /* caso, retorna-se 1, indicando que uma linha ou */
      /* coluna est completa.                          */
   if (linha < 9 && coluna < 9) {
         /* Verifica se o quadrado representado por 'linha'  */
         /* e 'coluna' est preenchido. Se for o caso,       */
         /* tenta-se preencher o quadrado na prxima coluna  */
         /* da mesma linha ou na prxima linha da mesma      */
         /* coluna. Se esse quadrado no existir, retorna-se */
         /* 1, informando que o quebra-cabea foi resolvido. */
      if(grade[linha][coluna]) {
            /* Se existir um quadrado na prxima coluna */
            /* desta linha, tenta-se preench-lo        */
         if (coluna + 1 < 9)
            return PreencheSudoku(grade, linha, coluna + 1);
            /* Se existir um quadrado na prxima linha */
            /* desta coluna, tenta-se preench-lo      */
         else if(linha + 1 < 9)
            return PreencheSudoku(grade, linha + 1, 0);
            /* No existe tal quadrado. Logo, retorna-se 1, */
            /* indicando que o quebra-cabea foi resolvido. */
         else
            return 1;
      } else {
            /* O quadrado representado por 'linha' */
            /* e 'coluna' no est preenchido.     */
         for(numero = 1; numero <= 9; ++numero) {
               /* Tenta preencher o quadrado */
               /* com o nmero corrente      */
            if(EstaDisponivel(grade, linha, coluna, numero)) {
               grade[linha][coluna] = numero;

                  /* Se existir um quadrado na prxima coluna */
                  /* desta linha, tenta-se preench-lo        */
               if(coluna + 1 < 9) {
                     /* Tenta preencher o quadrado. Se for  */
                     /* possvel, indica-se o sucesso; caso */
                     /* contrrio, torna-o vazio novamente  */
                     /* para provocar o retrocesso.         */
                  if(PreencheSudoku(grade, linha, coluna + 1))
                     return 1;
                  else
                     grade[linha][coluna] = 0;
                  /* Se existir um quadrado na prxima linha */
                  /* desta coluna, tenta-se preench-lo      */
               } else if(linha + 1 < 9) {
                     /* Tenta preencher o quadrado. Se for  */
                     /* possvel, indica-se o sucesso; caso */
                     /* contrrio, torna-o vazio novamente  */
                     /* para provocar o retrocesso.         */
                  if(PreencheSudoku(grade, linha + 1, 0))
                     return 1;
                  else
                     grade[linha][coluna] = 0;
                  /* O quadrado no existe tal. Retorna-se 1, */
                  /* indicando que o problema foi resolvido   */
               } else
                    return 1;
            }
         }
      }
      return 0;
   } else
      return 1; /* Problema resolvido */
}

/****
 *
 * ApresentaGrade(): Apresenta na tela uma grade que
 *                   representa o quebra-cabea Sudoku
 *
 * Parmetros:
 *      grade (entrada e sada) - array que representa a grade
 *
 * Retorno: Nada
 *
 * Observao: Esta funo no apresenta nada de instrutivo
 *             para estudantes de Estruturas de Dados. Ela
 *             apenas mostra como dados podem ser apresentados
 *             na tela com um pouco de embelezamento. Portanto,
 *             no desperdice seu tempo com esta funo.
 *
 ****/
void ApresentaGrade(int grade[][9])
{
      int i, j;

      printf("\n+-----+-----+-----+\n");

      for(i = 1; i < 10; ++i) {
         for(j = 1; j < 10; ++j)
            printf("|%d", grade[i - 1][j - 1]);

         printf("|\n");

         if (i%3 == 0)
            printf("+-----+-----+-----+\n");
      }
}

/****
 *
 * main(): Resolve um quebra-cabea Sudoku
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se o quebra-cabea for resolvido.
 *          1, em caso contrrio.
 *
 ****/
int main(void)
{
      /* Array que que representa a grade do quebra-cabea. */
      /* 0 significa um quadrado no preenchido.            */
   int grade[9][9]={ {8, 0, 0, 4, 0, 6, 0, 0, 7},
                     {0, 0, 0, 0, 0, 0, 4, 0, 0},
                     {0, 1, 0, 0, 0, 0, 6, 5, 0},
                     {5, 0, 9, 0, 3, 0, 7, 8, 0},
                     {0, 0, 0, 0, 7, 0, 0, 0, 0},
                     {0, 4, 8, 0, 2, 0, 1, 0, 3},
                     {0, 5, 2, 0, 0, 0, 0, 9, 0},
                     {0, 0, 1, 0, 0, 0, 0, 0, 0},
                     {3, 0, 0, 9, 0, 2, 0, 0, 5} };

      /* Apresenta o desafio a ser enfrentado */
   printf("\n>>> Quebra-cabeca a ser resolvido <<<\n");
   ApresentaGrade(grade);

      /* Verifica se o quebra-cabea foi resolvido */
   if(PreencheSudoku(grade, 0, 0)) { /* Foi resolvido */
         /* Apresenta o resultado */
      printf("\n>>> Quebra-cabeca resolvido <<<\n");
      ApresentaGrade(grade);
   } else { /* O quebra-cabea no foi resolvido     */
            /* provavelmente porque foi mal formulado */
      printf("\n\nEste Sudoku nao tem solucao\n\n");
      return 1;
   }

   return 0;
}

/****************** Resultado do Programa *****************

>>> Quebra-cabeca a ser resolvido <<<

+-----+-----+-----+
|8|0|0|4|0|6|0|0|7|
|0|0|0|0|0|0|4|0|0|
|0|1|0|0|0|0|6|5|0|
+-----+-----+-----+
|5|0|9|0|3|0|7|8|0|
|0|0|0|0|7|0|0|0|0|
|0|4|8|0|2|0|1|0|3|
+-----+-----+-----+
|0|5|2|0|0|0|0|9|0|
|0|0|1|0|0|0|0|0|0|
|3|0|0|9|0|2|0|0|5|
+-----+-----+-----+

>>> Quebra-cabeca resolvido <<<

+-----+-----+-----+
|8|3|5|4|1|6|9|2|7|
|2|9|6|8|5|7|4|3|1|
|4|1|7|2|9|3|6|5|8|
+-----+-----+-----+
|5|6|9|1|3|4|7|8|2|
|1|2|3|6|7|8|5|4|9|
|7|4|8|5|2|9|1|6|3|
+-----+-----+-----+
|6|5|2|7|8|1|3|9|4|
|9|8|1|3|4|5|2|7|6|
|3|7|4|9|6|2|8|1|5|
+-----+-----+-----+

**********************************************************/
