/****
 *
 * Ttulo: Gerenciando uma Lista de Compras com Sensibilidade
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2014
 * ltima modificao: 16/04/2014
 *
 * Descrio: Determina se um nmero inteiro positivo  primo
 *
 * Entrada: Valores inteiros positivos ou zero para encerrar o programa
 *
 * Sada: Uma mensagem indicando se o nmero  primo ou no
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Entrada e sada          */
#include <string.h>   /* Processamento de strings */
#include "Lista.h"    /* Interface do mdulo      */
#include "ASSEGURA.h" /* Macro ASSEGURA           */

/***************** Definies de Funes ******************/

/****
 *
 * EstaCheia(): Verifica se uma lista ordenada est cheia
 *
 * Parmetros: *lista (entrada) - a lista que ser testada
 *
 * Retorno: 1, se a lista estiver cheia; 0, em caso contrario
 *
 ****/
int EstaCheia(const tListaCompras *lista)
{
   return lista->nItens >= MAX_ITENS;
}

/****
 *
 * IniciaLista(): Inicia uma lista ordenada
 *
 * Parmetros: lista (sada) - a lista que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaLista(tListaCompras *lista)
{
   lista->nItens = 0;
}

/****
 *
 * Comprimento(): Calcula o comprimento de uma lista
 *
 * Parmetros: *lista (entrada) - a lista cujo comprimento ser calculado
 *
 * Retorno: O nmero de itens na lista
 *
 ****/
int Comprimento(const tListaCompras *lista)
{
   return lista->nItens; /* No h o que calcular */
}

/****
 *
 * ObtemItem(): Obtm o elemento de uma lista que se
 *                  encontra numa dada posio
 *
 * Parmetros:
 *     *lista (entrada) - a lista
 *      indice (entrada) - o elemento
 *
 * Retorno: o nmero de itens na lista
 *
 ****/
char *ObtemItem(const tListaCompras *lista, int pos)
{
   ASSEGURA( pos >= 0 && pos < lista->nItens, "Elemento inexistente" );

   return (char *) lista->itens[pos];
}

/****
 *
 * InsereItem(): Insere um novo elemento numa dada posio
 *                   de uma lista ordenada
 *
 * Parmetros:
 *      lista (entrada/sada) - lista na qual ser feita a insero
 *      elemento (entrada) - valor do elemento que ser inserido
 *      pos (entrada) - a posio de insero
 *
 * Retorno: Nada
 *
 ****/
void InsereItem(tListaCompras *lista, tItemCompra elemento, int pos)
{
   ASSEGURA(!EstaCheia(lista), "A lista esta cheia");

      /* Abre espao para o novo elemento */
   for (int i = lista->nItens; i >= pos; --i)
      strcpy( (char *)lista->itens[i + 1], (char *)lista->itens[i] );

      /* Insere o novo elemento */
   strcpy((char *)lista->itens[pos], (char *)elemento);

   ++lista->nItens; /* O tamanho da lista aumentou */
}

/****
 *
 * RemoveItem(): Remove um elemento numa dada posio
 *                   de uma lista ordenada
 *
 * Parmetros:
 *      *lista (entrada/sada) - lista onde ser feita a remoo
 *       pos (entrada) - posio do elemento que ser removido
 *
 * Retorno: Endereo do item removido
 *
 ****/
char *RemoveItem(tListaCompras *lista, int pos)
{
   int i;
   static tItemCompra itemRemovido;

   ASSEGURA(!EstaVazia(lista), "A lista esta' vazia");

   strcpy( (char *)itemRemovido, (char *)lista->itens[pos] );

      /* Remover um elemento significa mover cada elemento uma posio para   */
      /* trs a partir do prximo elemento adiante daquele que ser  removido */
   for (i = pos; i < lista->nItens - 1; i++)
      strcpy( (char *)lista->itens[i], (char *)lista->itens[i + 1] );

   --lista->nItens; /* O tamanho da lista diminuiu */

   return (char *)itemRemovido;
}

/****
 *
 * EstaVazia(): Verifica se uma lista ordenada est vazia
 *
 * Parmetros: *lista (entrada) - a lista que ser testada
 *
 * Retorno: 1, se a lista estiver vazia; 0, em caso contrario
 *
 ****/
int EstaVazia(const tListaCompras *lista)
{
   return lista->nItens == 0;
}

/****
 *
 * BuscaItem(): Procura um elemento numa lista que
 *                  apresenta um determinado valor
 *
 * Parmetros:
 *      *lista (entrada) - lista na qual ser efetuada a busca
 *       elemento (entrada) - valor do elemento que ser procurado
 *
 * Retorno: ndice do elemento se ele for encontrado; -1 em caso contrrio
 *
 ****/
int BuscaItem(const tListaCompras *lista, tItemCompra elemento)
{
   for (int i = 0; i < lista->nItens; ++i)
      if (!strcmp(lista->itens[i], elemento))
         return i; /* Elemento foi encontrado */

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * AcrescentaItem(): Acrescenta um novo elemento ao final de uma lista ordenada
 *
 * Parmetros:
 *      *lista (sada) - lista onde ser feita o acrscimo
 *       elemento (entrada) - valor do elemento que ser acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaItem(tListaCompras *lista, tItemCompra elemento)
{
   ASSEGURA(!EstaCheia(lista), "A lista esta cheia");

      /* Acrescenta um novo elemento ao final da lista */
   strcpy( (char *)lista->itens[lista->nItens], (char *)elemento );

   ++lista->nItens; /* O tamanho da lista aumentou */
}

/****
 *
 * AlteraItem(): Altera o valor de um elemento que se encontra num determinada
 *               posio de uma lista sem ordenao
 *
 * Parmetros:
 *      *lista (entrada e sada) - lista na qual o elemento se encontra
 *       elemento (entrada) - valor do elemento que ser acrescentado
 *
 * Retorno: 0, se a operao for bem sucedida; 1, em caso contrrio
 *
 ****/
void AlteraItem(tListaCompras *lista, tItemCompra valor, int pos)
{
   ASSEGURA( pos >= 0 && pos < lista->nItens, "Elemento inexistente" );

   strcpy((char *)lista->itens[pos], (char *)valor);
}
