/****
 *
 * Ttulo: Soluo Iterativa das Torres de Hani
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2015
 * ltima modificao: 16/04/2015
 *
 * Descrio: Resolve o Problema das Torres de Hani
 *            de modo iterativo
 *
 * Entrada: O nmero de discos do problema
 *
 * Sada: A sequncia de movimentao de discos a ser
 *        seguida para resolver o problema
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include "Pilha.h"   /* Interface do tipo pilha */
#include "leitura.h"

/***************** Definies de Funes ******************/

/****
 *
 * Troca(): Troca os valores de duas variveis do tipo char
 *
 * Parmetros:
 *      x1, x2 (entrada e sada) - valores que sero trocados
 *
 * Retorno: Nada
 *
 ****/
void Troca(char *x1, char *x2)
{
   char aux = *x1; /* Guarda o primeiro valor */

   *x1 = *x2;
   *x2 = aux;
}

/****
 *
 * TorresRecursivas(): Resolve o problema das Torres de Hani
 *                     recursivamente
 *
 * Parmetros:
 *      nDiscos (entrada) - nmero de discos
 *      hasteOrigem (entrada) - haste na qual os discos se
 *                              encontram inicialmente
 *      hasteDestino (entrada) - haste  qual os discos se
 *                               destinam
 *      hasteAuxiliar (entrada) - haste usada como auxiliar
 *
 * Retorno: Nada
 *
 ****/
void TorresRecursivas( int nDiscos, int hasteOrigem,
                       int hasteDestino, int hasteAuxiliar )
{
   if (nDiscos == 1){
         /* Passo 1 do algoritmo: escreve e encerra */
      printf( "Mova o disco 1 da haste %c para a haste %c\n",
              hasteOrigem, hasteDestino );
       return;
   }

      /* Passo 2 do algoritmo: move os n-1 discos */
      /* de A para B usando C como auxiliar       */
   TorresRecursivas( nDiscos - 1, hasteOrigem,
                     hasteAuxiliar, hasteDestino );

      /* Passo 3 do algoritmo: move ltimo disco de A para C */
   printf( "Mova o disco %d da haste %c para a haste %c\n",
           nDiscos, hasteOrigem, hasteDestino );

      /* Passo 4 do algoritmo: move os n-1 discos */
      /* de B para C usando A como auxiliar       */
   TorresRecursivas( nDiscos - 1, hasteAuxiliar,
                     hasteDestino,hasteOrigem );
}

/****
 *
 * TorresIterativas(): Resolve o problema das torres de
 *                     Hani iterativamente
 *
 * Parmetros:
 *      n (entrada) - nmero de discos
 *      hasteOrigem (entrada) - haste na qual os discos se
 *                              encontram inicialmente
 *      hasteDestino (entrada) - haste  qual os discos se
 *                               destinam
 *      hasteAuxiliar (entrada) - haste usada como auxiliar
 *
 * Retorno: Nada
 *
 ****/
void TorresIterativas( int n, int hasteOrigem, int hasteDestino,
                       int hasteAuxiliar )
{
   tEstado estado; /* Armazena um estado do problema */
   tPilhaIdx  pilha; /* Pilha de elementos do tipo tEstado */
                  /* que armazena estados do problema   */

   CriaPilhaIdx(&pilha);

      /* Armazena o estado inicial do problema, que so os  */
      /* valores dos parmetros reais recebidos pela funo */
   estado.n = n;
   estado.origem = hasteOrigem;
   estado.destino = hasteDestino;
   estado.aux = hasteAuxiliar;

      /* Transforma recurso em iterao substituindo    */
      /* chamadas recursivas em empilhamentos e retornos */
      /* de chamadas recursivas em desempilhamentos      */
   while (1) {
          /* Substitui a primera chamada recursiva */
      while (estado.n > 0) {
         EmpilhaIdx(&estado, &pilha);
         Troca(&estado.destino, &estado.aux);
         estado.n--;
      }

          /* Se a pilha ficou vazia, o */
          /* problema est resolvido   */
      if (PilhaIdxVazia(&pilha)) {
         break;
      } else {
            /* Desempilhamento corresponde a   */
            /* um retorno de chamada recursiva */
         estado = DesempilhaIdx(&pilha);
      }

         /* O ltimo estado desempilhado representa um   */
         /* movimento de discos que ser exibido na tela */
      printf( "Mova o disco %d da haste %c para a haste %c\n",
              estado.n, estado.origem, estado.destino );

          /* Substitui a segunda chamada recursiva */
      Troca(&estado.aux, &estado.origem);

      estado.n--;
   }
}

/****
 *
 * main(): Testa a funo TorresIterativas()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
  int nDiscos;

      /* Apresenta o programa ao usurio */
   printf( "\n\t>>> Este programa resolve o Problema das"
           "\n\t>>> Torres de Hanoi iterativamente\n" );

      /* L o numero de discos */
   printf("\n\t>>> Introduza o numero de discos: ");
   nDiscos = LeNaturalPositivo();

   putchar('\n'); /* Embelezamento */

   TorresIterativas(nDiscos, 'A', 'C', 'B');

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> Este programa resolve o Problema das
        >>> Torres de Hanoi iterativamente

        >>> Introduza o numero de discos: 4

Mova o disco 1 da haste A para a haste B
Mova o disco 2 da haste A para a haste C
Mova o disco 1 da haste B para a haste C
Mova o disco 3 da haste A para a haste B
Mova o disco 1 da haste C para a haste A
Mova o disco 2 da haste C para a haste B
Mova o disco 1 da haste A para a haste B
Mova o disco 4 da haste A para a haste C
Mova o disco 1 da haste B para a haste C
Mova o disco 2 da haste B para a haste A
Mova o disco 1 da haste C para a haste A
Mova o disco 3 da haste B para a haste C
Mova o disco 1 da haste A para a haste B
Mova o disco 2 da haste A para a haste C
Mova o disco 1 da haste B para a haste C

**********************************************************/
