/****
 *
 * Arquivo: Expressoes.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 10/04/2013
 * ltima modificao: 02/05/2013
 *
 * Descrio: Implementao de operaes de avaliao e
 *            converso de expresses
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>      /* Entrada e sada              */
#include <string.h>     /* Processamento de strings     */
#include <ctype.h>      /* Classificao de caracteres  */
#include <stdlib.h>     /* Alocao dinmica e strtol() */
#include "PilhaG.h"     /* Interface do mdulo PilhaG   */
#include "Expressoes.h" /* Interface deste mdulo       */
#include "ASSEGURA.h"   /* Macro ASSEGURA               */

/****************** Definies de Tipos *******************/

   /* Tipo de uma varivel que armazena */
   /* informaes sobre um operador     */
typedef struct {
           char operador; /* O operador  */
           int  aridade;  /* Sua aridade */
           int  pdp, /* Sua prioridade PDP */
                pc;  /* Sua prioridade PC  */
        } tOperador;

   /* Tipos de prioridade:                   */
   /*       PDP = prioridade dentro da pilha */
   /*       PC = prioridade de chegada       */
typedef enum {PDP, PC} tPrioridade;

/****************** Variveis de Arquivo ******************/

   /* Array que armazena informaes sobre os operadores */
static const tOperador operadores[] = { {'_', 1, 3, 4},
                                        {'*', 2, 2, 2},
                                        {'/', 2, 2, 2},
                                        {'+', 2, 1, 1},
                                        {'-', 2, 1, 1},
                                        {'(', 0, 0, 4}
                                      };
   /* Tamanho do array operadores[] */
static const int nOperadores  = sizeof(operadores) /
                                sizeof(operadores[0]);

/************** Definies de Funes Locais **************/

/****
 *
 * Prioridade(): Obtm a prioridade (PDP/PC) de um operador
 *
 * Parmetros:
 *      operador (entrada) - o operador
 *      tp (entrada) - tipo de prioridade (i.e., PDP ou PC)
 *
 * Retorno: A prioridade PDP ou PC do operador se ele for
 *          encontrado. Caso contrrio, um valor negativo.
 *
 * Observao: O programa ser abortado se o tipo de
 *             prioridade for desconhecido
 *
 ****/
static int Prioridade(int operador, tPrioridade tp)
{
   int i;

      /* O programa ser abortado se o tipo */
      /* de prioridade no for conhecido    */
   ASSEGURA( tp == PDP || tp == PC,
             "Erro: Tipo de prioridade desconhecida");

      /* Procura o operador no array */
   for (i = 0; i < nOperadores; ++i) {
      if (operadores[i].operador == operador) {
            /* O operador foi encontrado. Retorna */
            /* a prioridade do tipo especificado. */
         return tp == PDP ? operadores[i].pdp
                          : operadores[i].pc;
      }
   }

      /* O operador no foi encontrado */
   return -1;
}

/****
 *
 * Aridade(): Obtm a aridade de um operador
 *
 * Parmetros:
 *      operador (entrada) - o operador
 *
 * Retorno: A aridade do operador se ele for encontrado.
 *          Caso contrrio, um valor negativo
 *
 ****/
static int Aridade(int operador)
{
   int i;

      /* Procura o operador no array */
   for (i = 0; i < nOperadores; ++i) {
      if (operadores[i].operador == operador) {
            /* O operador foi encontrado. */
            /* Retorna sua aridade.       */
         return operadores[i].aridade;
      }
   }

      /* O operador no foi encontrado */
   return -1;
}

/****
 *
 * EhOperador(): Verifica se um caractere  um operador
 *
 * Parmetros:
 *      c (entrada) - o caractere a ser testado
 *
 * Retorno: 1, se o caractere for um operador
 *          0, em caso contrrio
 *
 ****/
static int EhOperador(int c)
{
   static const char *const strOperadores = "_*/+-()";

      /* Se strchr() retornar um ponteiro   */
      /* diferente de NULL, o caractere faz */
      /* parte do string 'strOperadores'    */
   return strchr(strOperadores, c) != NULL;
}

/****
 *
 * ProximoToken(): Obtm o prximo token de uma expresso
 *
 * Parmetros:
 *      expressao (entrada) - a expresso
 *      *fim (sada) - ponteiro para o ltimo caractere do
 *                     token
 *
 * Retorno: Endereo inicial do token, se ele existir;
 *          NULL, em caso contrrio
 *
 * Observao: O programa ser abortado se for encontrado um
 *             caractere que no seja dgito fazendo parte
 *             de um operando
 *
 ****/
static char *ProximoToken(const char *expressao, char **fim)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, *inicio;


   if (expressao) { /* A procura iniciar em 'expressao' */
      s = (char *)expressao;
   } else if (proximoToken) { /* A procura iniciar */
      s = proximoToken;       /* em 'proximoToken'  */
   } else {/* 'expressao' e 'proximoToken' so ambos NULL */
      return NULL; /* No h token */
   }

      /* O lao a seguir salta eventuais espaos em branco*/
      /* que porventura se encontrem no incio do string. */
      /* Ele encerra quando for encontrado um caractere   */
      /* que no seja espao em branco ou o caractere     */
      /* terminal do string apontado por s.               */
   while (*s && isspace(*s)) {
      ++s; /* Salta mais um espao */
   }

      /* Se o lao anterior encerrou porque foi  */
      /* encontrado o caractere terminal, no h */
      /* mais tokens no string                   */
   if (!*s) {
      proximoToken = NULL; /* No haver prximo token */

      return NULL; /* Nem h agora */
   }

      /* Guarda o incio do token corrente */
   inicio = s;

   /*******************************************/
   /*                                         */
   /* Como, por simplicidade, parnteses      */
   /* so considerados operadores, s h dois */
   /* tipos de tokens: operadores e operandos */
   /* numricos inteiros                      */
   /*                                         */
   /*******************************************/

      /* Verifica se o prximo caractere  um */
      /* operador (incluindo parnteses)      */
   if (EhOperador(*s)) { /* Trata-se de um operador */
      *fim = inicio; /* Operador s usa um caractere */

         /* O proximo token iniciar no prximo caractere */
      proximoToken = inicio + 1;

      return inicio; /* Servio concludo */
   }

      /* Se o caractere corrente no era um operador, */
      /* ele deve ser um dgito. Se ele no for um    */
      /* dgito, o programa ser abortado.            */
   ASSEGURA( isdigit(*s),
             "Erro em ProximoToken(): caractere "
             "espurio encontrado" );

      /* Enquanto o caractere corrente for um dgito, */
      /* faz s apontar para o prximo caractere       */
   while (isdigit(*s)) {
      ++s; /* Passa para o prximo caractere */
   }

      /* Neste ponto, s aponta para o primeiro */
      /* caractere que no  dgito. Este ser */
      /* o incio do prximo token.            */
   proximoToken = s;

      /* Faz *fim apontar para o ltimo dgito encontrado */
   *fim = s - 1;

      /* Retorna o endereo inicial do token corrente */
   return inicio;
}

/****
 *
 * ObtemOperando(): Obtm o operando inteiro supostamente
 *                  representado pelo token especificado
 *
 * Parmetros:
 *      token (entrada) - string que representa o token
 *
 * Retorno: O token convertido em inteiro
 *
 * Observao: O programa ser abortado se o token no
 *             representar um nmero inteiro positivo ou
 *             se ocorrer overflow durante a converso.
 *
 ****/
static int ObtemOperando(const char *token)
{
   int digito, /* Armazenar cada dgito do nmero */
       numero = 0; /* Armazenar o resultado da converso */

      /* O lao encerra quando o array for  */
      /* totalmente explorado ou se ocorrer */
      /* algum erro durante a converso     */
   while (*token) {
         /* Se for encontrado um caractere que no */
         /*  dgito, o programa ser abortado     */
      ASSEGURA( isdigit(*token),
                "Erro em ObtemOperando(): "
                "Caractere deveria ser digito" );

         /* Obtm nmero equivalente ao dgito encontrado */
      digito = *token - '0';

         /* Atualiza o valor numrico. A base */
         /* numrica utilizada  10.          */
      numero = numero*10 + digito;

         /* O nmero deve ser positivo. Portanto, se ele */
         /* assumir um valor negativo, ocorreu overflow. */
      ASSEGURA(numero >= 0, "\a\t>>> Overflow!!! <<<");

      token++; /* Passa para o prximo caractere */
   }

   return numero; /* A converso foi bem sucedida */
}

/****
 *
 * AnexaEmExpressao(): Anexa um novo termo a uma expresso
 *
 * Parmetros:
 *     *expressao (entrada/sada) - array contendo um string
 *                                  que representa a expresso
 *     *tam (entrada/sada) - tamanho do array que contm a
 *                            expresso
 *      termo (entrada) - array de caracteres contendo o
 *                        novo termo (no  string)
 *      tamTermo (entrada) - nmero de caracteres do termo
 *
 * Retorno: Nada
 *
 * Observao: O programa ser abortado se ocorrer falha de
 *             alocao dinmica de memria
 *
 ****/
static void AnexaEmExpressao( char **expressao, int *tam,
                              char *termo, int tamTermo )
{
   char *p, *str;
   int   tamExpressao, i;

   tamExpressao = strlen(*expressao) + /* Tamanho corrente*/
                                       /* da expresso    */
                  tamTermo + /* Tamanho do novo termo */
                  2;   /* Caractere terminal    */
                       /* mais espao em branco */

      /* Se o nmero de caracteres necessrios for */
      /* maior do que o tamanho do array, tenta-se */
      /* aumentar o tamanho do array               */
   if (tamExpressao > *tam) {
         /* Se a realocao no for     */
         /* possvel, aborta o programa */
      ASSEGURA( (p = realloc(*expressao, tamExpressao)),
                "Erro em AnexaEmExpressao(): "
                "Realocacao nao foi possivel");

         /* Atualiza o tamanho do array */
      *tam = tamExpressao;

         /* Atualiza ponteiro que aponta para a expresso */
      *expressao = p;
   }

      /* O uso do ponteiro 'str' no  essencial, */
      /* mas facilita a escrita e a legibilidade  */
      /* das instrues seguintes                 */
   str = *expressao;

      /* Faz str apontar para o caractere terminal */
      /* do string que representa a expresso      */
   str = str + strlen(str);

      /* Copia os caracteres do novo termo para a       */
      /* expresso, de modo que o primeiro caractere do */
      /* termo seja armazenado no lugar do caractere    */
      /* terminal da expresso e os demais caracteres   */
      /* sejam armazenados nas posies subsequentes    */
   for (i = 1; i <= tamTermo; ++i) {
      *str++ = *termo++;
   }

   *str++ = ' '; /* Armazena um espao em branco */
   *str = '\0'; /* Termina a nova expresso */
}

/****
 *
 * ResultadoOperacao(): Calcula o resultado de uma operao
 *                      aritmtica simples envolvendo um
 *                      operador unrio e seu nico operando
 *                      ou um operador binrio e seus dois
 *                      operandos
 *
 * Parmetros:
 *      operador (entrada) - operador da expresso
 *      op1 (entrada) - primeiro operando
 *      op2 (entrada) - segundo operando
 *
 * Retorno: Nada
 *
 * Observaes:
 *      1. Se o operador for unrio, seu operando  op1.
 *         Neste caso, op2 no  levado em considerao.
 *
 *      2. Se o operador for de diviso, op1  o dividendo
 *         e op2  o divisor.
 *
 *      3. Se o operador for de subtrao, op1  o minuendo
 *         e op2  o subtraendo.
 *
 *      4. O programa ser abortado se o operador for
 *         indeterminado.
 *
 ****/
static int ResultadoOperacao(int operador, int op1, int op2)
{
      /* Calcula e retorna o resultado da operao */
   switch (operador) {
      case '_': /* Inverso de sinal */
         return -op1;
         break;
      case '*': /* Multiplicao */
         return op1 * op2;
         break;
      case '/': /* Diviso */
         return op1 / op2;
         break;
      case '+': /* Soma */
         return op1 + op2;
         break;
      case '-': /* Subtrao */
         return op1 - op2;
         break;
      default: /* Operador indeterminado */
         ASSEGURA(0, "Nao foi possivel encontrar operador\n");
   }

      /* A instruo a seguir nunca ser executada, */
      /* mas torna a funo compatvel com C++      */
   return 0;
}

/************* Definies de Funes Globais **************/

/****
 *
 * ExpressaoSufixa(): Obtm a expresso sufixa correspondente
 *                    a uma expresso infixa
 *
 * Parmetros:
 *      infixa (entrada) - string que representa a expresso
 *                         infixa
 *
 * Retorno: Endereo da expresso sufixa resultante
 *
 * Observao: O programa ser abortado se o array que
 *             armazenar a expresso sufixa no puder
 *             ser alocado ou se houver um fecha-parnteses
 *             sem o respectivo abre-parnteses na expresso
 *             infixa.
 *
 ****/
char *ExpressaoSufixa(const char *infixa)
{
   tPilhaGen pilha; /* Pilha que armazenar operadores */
                    /* e abre-parnteses               */
   char     *inicio, /* Apontar para o incio de um token*/
            *fim; /* Apontar para o final de um token */
   char      e; /* Elemento da pilha */
   char     *sufixa; /* Ponteiro para a expresso sufixa */
   int       tamSufixa; /* Tamanho da expresso sufixa */

      /* O tamanho inicial da expresso sufixa    */
      /* estimado como sendo igual ao da expresso */
      /* infixa. Se for necessrio, este tamanho   */
      /* ser devidamente aumentado.               */
   tamSufixa = strlen(infixa) + 1;

      /* Tenta alocar espao para  */
      /* conter a expresso sufixa */
   ASSEGURA( sufixa = calloc(tamSufixa, sizeof(char)),
             "Erro em ExpressaoSufixa(): Impossivel "
             "alocar espaco para a expressao" );

      /* Inicia a pilha que armazenar operadores */
   CriaPilhaG(&pilha, sizeof(char));

      /* Obtm o primeiro token */
   inicio = ProximoToken(infixa, &fim);

      /* O lao a seguir encerra quando no     */
      /* houver mais tokens na expresso infixa */
   while (inicio) {
      if (*inicio == ')') {
            /* O token corrente  um fecha-parnteses.  */
            /* Ento, a pilha deve ser desempilhada at */
            /* o correspondente abre-parnteses.        */
         while (!PilhaVaziaG(&pilha) &&
               *(char *)(ElementoTopoG(&pilha, &e)) != '('){
            DesempilhaG(&pilha, &e);
            AnexaEmExpressao(&sufixa, &tamSufixa, &e, 1);
         }

            /* Neste ponto, se a pilha estiver vazia ou  */
            /* se o prximo token desempilhado no for   */
            /* abre-parnteses, o programa ser abortado */
         ASSEGURA( !PilhaVaziaG(&pilha) &&
                   *(char *) DesempilhaG(&pilha, &e) == '(',
                   "Erro: Nao foi encontrado '('" );
      } else if (EhOperador(*inicio)) {
            /* O token corrente  um operador ou abre-    */
            /* parnteses. Ento, enquanto a pilha no    */
            /* estiver vazia e a prioridade PDP do token  */
            /* no topo da pilha for maior ou igual que a  */
            /* prioridade PC do token corrente, o token   */
            /* encontrado no topo da pilha  desempilhado */
            /* e anexado  expresso sufixa.              */
         while (!PilhaVaziaG(&pilha) &&
               Prioridade(*(char *)ElementoTopoG(&pilha,&e),
                          PDP) >=
                Prioridade(*inicio, PC) ) {
            DesempilhaG(&pilha, &e);
            AnexaEmExpressao(&sufixa, &tamSufixa, &e, 1);
         }

            /* Empilha o token corrente. NB: Apesar de se */
            /* estar passando um ponteiro para um string  */
            /* como segundo parmetro, apenas o primeiro  */
            /* caractere deste string ser empilhado.     */
         EmpilhaG(&pilha, inicio);
      } else {
            /* O token corrente  um operando, que     */
            /* imediatamente anexado  expresso sufixa */
         AnexaEmExpressao( &sufixa, &tamSufixa, inicio,
                           fim - inicio + 1 );
      }

         /* Obtm o prximo token */
      inicio = ProximoToken(NULL, &fim);
   }

      /* Neste ponto, a pilha contm apenas */
      /* operadores que sero desempilhados */
      /* e anexados  expresso sufixa      */
   while (!PilhaVaziaG(&pilha)) {
      DesempilhaG(&pilha, &e);
      AnexaEmExpressao(&sufixa, &tamSufixa, &e, 1);
   }

      /* Libera espao ocupado pela pilha */
   DestroiPilhaG(&pilha);

   return sufixa; /* Servio completo */
}

/****
 *
 * AvalExpressaoSufixa(): Avalia uma expresso sufixa
 *
 * Parmetros:
 *      sufixa (entrada/sada) - string representando a
 *                               expresso que ser avaliada
 *
 * Retorno: O resultado da avaliao
 *
 * Observao: O programa ser abortado se a pilha no
 *             estiver vazia ao final da avaliao ou se
 *             a aridade de um operador no for vlida
 *
 ****/
int AvalExpressaoSufixa(char *sufixa)
{
   tPilhaGen pilha; /* Pilha que armazenar operandos */
   int       op1, op2, /* Dois operandos de uma expresso */
             resultado, /* Resultado de uma expresso */
             ari; /* Aridade de um operador */
   char     *token, /* Um token da expresso sufixa */
            *final; /* Apontar para o caractere que termina */
                    /* uma converso de string em inteiro    */

      /* Inicia a pilha que armazenar operandos */
   CriaPilhaG(&pilha, sizeof(int));

      /* Obtm o primeiro token. Espao em branco */
      /*  o nico separador de tokens.           */
   token = strtok(sufixa, " ");

      /* O lao a seguir encerra quando no     */
      /* houver mais tokens na expresso sufixa */
   while (token) {
      if (EhOperador(*token)) { /* Token  operador */
            /* Obtm a aridade do operador */
         ari = Aridade(*token);

            /* A aridade deve ser 1 ou 2. Se no for */
            /* o caso, o programa ser abortado.     */
         ASSEGURA( (ari == 1 || ari == 2),
                   "Erro em AvalExpressaoSufixa(): "
                   "Aridade deveria ser 1 ou 2" );

         if (ari == 1) {
               /* O operador  unrio. Ento,       */
               /* desempilha-se um operando e       */
               /* obtm-se o resultado da expresso */
            DesempilhaG(&pilha, &op1);

               /* O terceiro parmetro da chamada de    */
               /* ResultadoOperacao() pode ter qualquer */
               /* valor, uma vez que ele no ser usado */
            resultado = ResultadoOperacao(*token, op1, 0);
         } else {
               /* O operador  binrio. Ento,      */
               /* desempilham-se dois operandos e   */
               /* obtm-se o resultado da expresso */
            DesempilhaG(&pilha, &op2);
            DesempilhaG(&pilha, &op1);
            resultado = ResultadoOperacao(*token, op1, op2);
         }

            /* O resultado da operao deve ser empilhado */
         EmpilhaG(&pilha, &resultado);
      } else {
            /*****************************************/
            /* O token corrente  um string que deve */
            /* representar um operando inteiro       */
            /*****************************************/

            /* Obtm o valor inteiro    */
            /* representado pelo string */
         op1 = strtol(token, &final, 10);

         ASSEGURA(!*final, "Conversao de operando falhou");

            /* Empilha o operando inteiro */
         EmpilhaG(&pilha, &op1);
      }

         /* Obtm o prximo token da expresso sufixa */
      token = strtok(NULL, " ");
   }

      /* O resultado da avaliao est no topo da pilha */
   DesempilhaG(&pilha, &resultado);

      /* Teste de consistencia: neste ponto, */
      /* a pilha deve estar vazia            */
   ASSEGURA( PilhaVaziaG(&pilha),
             "Erro em AvalExpressaoSufixa(): "
             "A pilha deveria estar vazia" );

      /* Libera espao ocupado pela pilha */
   DestroiPilhaG(&pilha);

   return resultado; /* Servio completo */
}

/****
 *
 * AvalExpressaoInfixa(): Avalia uma expresso infixa
 *
 * Parmetros:
 *      infixa (entrada) - string representando a expresso
 *                         infixa que ser avaliada
 *
 * Retorno: O resultado da avaliao da expresso
 *
 ****/
int AvalExpressaoInfixa(const char *infixa)
{
   char *sufixa; /* Apontar para a expresso */
                 /* sufixa correspondente     */
   int   resultado; /* Armazenar o resultado da */
                    /* avaliao da expresso    */

      /* Obtm a expresso sufixa correspondente */
   sufixa = ExpressaoSufixa(infixa);

      /* Obtm resultado da avaliao da expresso sufixa */
   resultado = AvalExpressaoSufixa(sufixa);

      /* O espao ocupado pela expresso sufixa foi  */
      /* alocado dinamicamente e  preciso liber-lo */
   free(sufixa);

   return resultado;
}
