/****
 *
 * Ttulo: Codificao de Huffman
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2015
 * ltima modificao: 16/04/2015
 *
 * Descrio: Implementa a Codificao de Huffman
 *
 * Entrada: Um string contendo apenas letras (caso contrrio,
 *          o programa  abortado)
 *
 * Sada: A codificao de Huffman para cada caractere que
 *        compe um string

 * OBSERVAO: Neste programa, os bits 0 e 1 so representados
 *             pelos caracteres '0' e '1'. Desse modo,
 *             obviamente, a codificao das letras ocupa mais
 *             do que as prprias letras. Na prtica, os bits
 *             reais da codificao so obtidos por meio de
 *             processamento de baixo nvel, o que est
 *             alm do escopo deste livro. Aqui, o objetivo
 *              demonstrar o raciocnio que norteia o
 *             algoritmo de Huffman e o uso de rvore binria
 *             e lista encadeada na implementao desse algoritmo.
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>   /* Entrada e sada              */
#include <string.h>  /* Processamento de strings     */
#include <stdlib.h>  /* Alocao dinmica            */
#include <ctype.h>   /* Classificao de caracteres  */

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              fprintf(stderr, "\n%s", msg); \
                              exit(1); \
                           }

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

/****************** Definies de Tipos *******************/

   /* Tipo do campo de informao dos */
   /* ns da rvore de codificao    */
typedef struct {
          char   letra;
          double frequencia;
        } tLetraFreq;

   /* Tipo de cada n e tipo de ponteiro */
   /* para n da rvore de codificao   */
typedef struct noArvore {
          tLetraFreq        letraFreq;
          struct noArvore  *filhoEsquerda;
          struct noArvore  *filhoDireita;
        } tNoArvore, *tArvore;

   /* Tipo de n e tipo de ponteiro  */
   /* para n da lista encadeada que */
   /* armazena os ns da rvore de   */
   /* codificao temporariamente    */
typedef struct noLista {
           tNoArvore       conteudo;
           struct noLista *proximo;
        } tNoLista, *tLista;

   /* Tipo de n e tipo de ponteiro  */
   /* para n da lista encadeada que */
   /* armazena os cdigos de letras  */
typedef struct noCodigo {
           char             letra;
           char            *codigo;
           struct noCodigo *proximo;
        } tNoCodigo, *tListaDeCodigos;

/************************* Aluses ************************/

extern int LimpaBuffer(void);
extern int LeCaractere(void);
extern int LeOpcao(const char *opcoes);
extern char *LeLinhaIlimitada(int *tam, FILE *stream);

extern int ComprimentoLE(tLista lista);
extern tLista CriaListaEncadeadaOrd(void);
extern tNoLista *CriaNo(int letra);
extern void InsereNoEmOrdem(tLista *lista, tNoLista *pNo);
extern tLista RemovePrimeiroNo(tLista *lista);

extern tArvore Codifica(tLista *aLista);
extern void DestroiArvoreBin(tArvore arvore);
extern void DestroiListaDeCodigos(tListaDeCodigos *lista);
extern void InsereEmListaDeCodigos( tListaDeCodigos *lista,
                                    int letra,
                                    char *codigo, int n );
extern void ColetaCodigos( tArvore raiz, char codigo[],
                           int n,
                           tListaDeCodigos *listaDeCodigos );
extern void ExibeCodigos(tListaDeCodigos lista);
extern void ExibeCodificacao(const char *string,
                             tListaDeCodigos listaDeCodigos);

extern double FrequenciasDeLetras(int aLetra);
extern char *RemoveBrancos(char *str);
extern void ApresentaMenu(void);

/***************** Definies de Funes ******************/

/******************* Funes de Leitura *******************/

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeCaractere(): L um caractere via teclado e remove os
 *                caracteres remanescentes no buffer
 *
 * Parmetros: Nenhum
 *
 * Retorno: O caractere lido
 *
 * Observao: Esta funo deixa o buffer limpo
 *
 ****/
int LeCaractere(void)
{
   int c, /* Armazenar cada caractere lido */
       nResto = 0; /* Nmero de caracteres */
                   /* excedentes           */

   /* Volta para c se ocorrer erro de leitura */
inicio:
   c = getchar(); /* L o caractere digitado */

      /* Verifica se getchar() retornou EOF */
   if (c == EOF) {
      printf( "\a\n\t>>>Erro de leitura. "
              "Tente novamente\n\t> " );
      goto inicio; /* No faz mal algum */
   }

      /* Pelo menos '\n' se encontra no buffer */
   nResto = LimpaBuffer();

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf( "\t>>> %d caracteres foram descartados\n",
              nResto );
   }

   return c; /* Retorna o caractere lido */
}

/****
 *
 * LeOpcao(): L, via teclado, um caractere que deve fazer
 *            parte de um string para ser considerado vlido
 *
 * Parmetros:
 *      opcoes (entrada) - string contendo os caracteres
 *                         vlidos
 *
 * Retorno: Um caractere vlido
 *
 * Observaes:
 *      1. Esta funo deixa o buffer virgem novamente
 *      2. Esta funo  denominada 'LeOpcao' porque o
 *         caractere lido e validado  restrito a um
 *         conjunto de caracteres (opes) representado
 *         pelo string recebido como parmetro
 *
 ****/
int LeOpcao(const char *opcoes)
{
   int op; /* Opo (caractere) escolhida pelo usurio */

      /* Enquanto o usurio no escolher uma opo */
      /* vlida, o lao a seguir no encerra       */
   while (1) {
         /* L o caractere digitado */
      op = LeCaractere();

         /* Verifica se o caractere  vlido */
      if (strchr(opcoes, op)) {
         break; /*  */
      } else { /* No  */
         printf( "\a\n\t>>> Opcao incorreta. "
                "Tente novamente\n\t> " );
      }
   }

      /* Certamente, a opo retornada  vlida */
   return op;
}

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor que i*/
         /* poderia assumir deveria ser tamanho - 1. Mas, como  */
         /* ao final, o caractere '\0' dever ser inserido,     */
         /* limita-se o valor de i a tamanho - 2.               */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontando pelo parmetro */
      /* 'tam', se ele no for NULL                */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****** Processamento de Listas Encadeadas Ordenadas ******/

/****
 *
 * ComprimentoLE(): Calcula o nmero de ns de uma lista
 *                  encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para o primeiro n da lista
 *                       cujo comprimento ser calculado
 *
 * Retorno: O nmero de ns da referida lista
 *
 * Observao: Esta funo independe do tipo de contedo efetivo
 *             armazenado em cada n da lista
 *
 ****/
int ComprimentoLE(tLista lista)
{
   int tamanho = 0; /* Armazena o nmero de ns da lista */

      /* Acessa cada n da lista e conta */
      /* quantos ns so acessados       */
   while (lista) {
      ++tamanho; /* Mais um n foi encontrado */
      lista = lista->proximo; /* Passa para o prximo n */
   }

   return tamanho;
}

/****
 *
 * CriaListaEncadeadaOrd(): Cria uma lista contendo ns de uma
 *                          rvore cujos ns armazenam caracteres
 *                          e suas respectivas frequncias
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da lista criada.
 *
 ****/
tLista CriaListaEncadeadaOrd(void)
{
   tLista    aLista; /* A lista que ser criada */

   aLista = NULL; /* A lista deve ser iniciada com NULL */

      /*                                  */
   for (int i = 'A'; i < 'Z'; ++i) {
      if (isalpha(i)) {
         InsereNoEmOrdem(&aLista, CriaNo(i));
      }
   }

   for (int i = 'a'; i < 'z'; ++i) {
      if (isalpha(i)) {
         InsereNoEmOrdem(&aLista, CriaNo(i));
      }
   }

   InsereNoEmOrdem(&aLista, CriaNo(' '));
   InsereNoEmOrdem(&aLista, CriaNo('-'));

   return aLista; /* Retorna o endereo inicial da lista */
}

/****
 *
 * CriaNo(): Insere um novo n numa lista encadeada
 *                    de modo que ela seja ORDENADA pelo campo
 *                    'frequencia' do contedo de cada n
 *
 * Parmetros:
 *     *lista (entrada e sada) - ponteiro para a lista na qual
 *                                ser efetuada a insero
 *      pNo (entrada e sada) - ponteiro para o n que ser
 *                              inserido
 *
 * Retorno: Nada
 *
 ****/
tNoLista *CriaNo(int letra)
{
   tNoLista *ptrNovoNo; /* Apontar para o novo n alocado  */

      /* Tenta alocar um novo n. Se no for */
      /* possvel, aborta o programa.        */
   ASSEGURA( ptrNovoNo = malloc(sizeof(tNoLista)),
             "Nao foi possvel alocar no da lista" );

      /* Armazena no novo n a letra */
      /* e sua frequncia            */
   ptrNovoNo->conteudo.letraFreq.letra = letra;
   ptrNovoNo->conteudo.letraFreq.frequencia =
                                FrequenciasDeLetras(letra);

      /* Quando FrequenciasDeLetras() retorna um */
      /* valor negativo, o caracter no possui   */
      /* frequncia especificada. Nesse caso, o  */
      /* programa  abortado                     */
   ASSEGURA( ptrNovoNo->conteudo.letraFreq.frequencia >= 0,
             "Encontrado caractere sem frequencia" );

      /* Inicialmente, a lista s contm folhas da rvore */
   ptrNovoNo->conteudo.filhoEsquerda = NULL;
   ptrNovoNo->conteudo.filhoDireita = NULL;

   return ptrNovoNo;
}

/****
 *
 * InsereNoEmOrdem(): Insere um novo n numa lista encadeada
 *                    de modo que ela seja ORDENADA pelo campo
 *                    'frequencia' do contedo de cada n
 *
 * Parmetros:
 *     *lista (entrada e sada) - ponteiro para a lista na qual
 *                                ser efetuada a insero
 *      pNo (entrada e sada) - ponteiro para o n que ser
 *                              inserido
 *
 * Retorno: Nada
 *
 ****/
void InsereNoEmOrdem(tLista *lista, tNoLista *pNo)
{
   tLista    p = *lista, /* p aponta para o n corrente     */
             q = NULL;   /* q aponta para o n anterior a p */

      /* Se a lista estiver vazia, basta adicionar */
      /* o n ao incio da lista e retornar        */
   if (!*lista) {
      *lista = pNo; /* Incio da lista apontar */
                    /* para novo n             */
      (*lista)->proximo = NULL; /* O novo n  o nico da lista */
      return; /* Servio completo */
   }

         /****************************************************/
         /* A lista tem pelo menos um n. Nesse caso, a      */
         /* posio de insero  aquela do primeiro n cujo */
         /* campo 'frequencia' seja maior do que aquele do   */
         /* n a ser inserido. Se tal n no for encontrado, */
         /* o novo n ser o ltimo da lista.                */
         /****************************************************/

      /* Procura o local de insero. Este no  o modo */
      /* mais adequado para comparar nmeros de ponto   */
      /* flutuante (i.e., as frequncias), mas aqui no */
      /* se requer tanta preciso.                      */
   while ( p &&
           p->conteudo.letraFreq.frequencia <=
                 pNo->conteudo.letraFreq.frequencia ){
      q = p; /* q passa a apontar para o n corrente */
      p = p->proximo; /* p passa a apontar para o prximo n */
   }

      /**********************************************************/
      /* Neste instante, p aponta para o local da insero e q  */
      /* aponta para o n imediatamente antes desse local. Se   */
      /* p for NULL, o novo n ser o ltimo da lista, mas isso */
      /* no constitui um caso especial a ser tratado  parte.  */
      /**********************************************************/

      /* O novo n apontar para o n apontado por p */
   pNo->proximo = p;

      /* Verifica se p aponta para o primeiro n da lista pois */
      /* insero no incio deve ser tratada separadamente     */
   if (p == *lista) { /* Insero ser no incio da lista */
      *lista = pNo;/* Insero no incio  especial */
   } else {
         /* O n anterior quele apontado */
         /* por p apontar para o novo n */
      q->proximo = pNo;
   }
}

/****
 *
 * RemovePrimeiroNo(): Remove o primeiro n de uma lista
 *                     encadeada
 *
 * Parmetros:
 *     *lista (entrada e sada) - ponteiro para o primeiro n
 *                                da lista na qual ser feita
 *                                a remoo
 *
 * Retorno: O endereo do n removido

 * Observaes:
 *    1. O n  removido da lista, mas seu espao no  liberado
 *    2. Se a lista estiver vazia, o programa ser abortado
 *
 ****/
tLista RemovePrimeiroNo(tLista *lista)
{
   tLista p = *lista; /* Guarda endereo do primeiro n */

      /* Se a lista estiver vazia, o programa  abortado */
   ASSEGURA( *lista,
             "Erro: Tentativa de remocao em lista vazia" );

      /* Efetua a remoo */
   (*lista) = (*lista)->proximo; /* Altera incio da lista */

   return p;
}

/***************** Codificao de Huffman *****************/

/****
 *
 * Codifica(): Cria uma rvore binria de codificao
 *             seguindo o algoritmo de Huffman
 *
 * Parmetros:
 *     *aLista (entrada e sada) - ponteior para a lista que
 *                                 armazena temporariamente
 *                                 os ns da rvore binria
 *                                 que representa uma
 *                                 codificao de Huffman
 *
 * Retorno: Endereo da raiz da rvore binria que representa
 *          a codificao de Huffman criada
 *
 * Observao: Esta funo aborta o programa quando no
 *             consegue alocar um n da rvore
 *
 ****/
tArvore Codifica(tLista *aLista)
{
   tNoLista *no1, /* Ponteiro para o n com a  */
                  /* menor frequncia na lista */
            *no2, /* Ponteiro para o n com a segunda */
                  /* menor frequncia na lista        */
            *noNovo; /* Ponteiro para um novo n da  */
                     /* rvore a ser criado. Esse n */
                     /* ter no1 e no2 como filhos.  */

      /***************************************************/
      /*                                                 */
      /* O lao while a seguir  a essncia do algoritmo */
      /* de Huffman. Ele remove da lista os dois ns com */
      /* menor frequncia, cria um novo n que tem como  */
      /* frequncia a soma das frequncias dos ns       */
      /* removidos e esses ns como filhos. O novo n   */
      /* inserido em ordem crescente de frequncia na    */
      /* lista. Quando resta apenas um n na lista, esse */
      /* ltimo n  a raiz da rvore de codificao.    */
      /*                                                 */
      /***************************************************/

   while (*aLista) {
         /* Remove o n com a menor frequncia da lista */
      no1 = RemovePrimeiroNo(aLista);

         /* Se a lista ficou vazia, o ltimo n removido  */
         /* a raiz da rvore que representa a codificao  */
      if (!*aLista) {
         return &no1->conteudo; /* Codificao terminada */
      }

         /* Remove o n com a segunda */
         /* menor frequncia da lista */
      no2 = RemovePrimeiroNo(aLista);

         /* Tenta alocar um novo n */
      ASSEGURA( noNovo = malloc(sizeof(tNoLista)),
                "Nao foi possvel alocar no da lista" );

         /* Armazena a frequncia do novo n. O contedo */
         /* do campo 'letra' no tem importncia porque  */
         /* este no  um n terminal.                   */
      noNovo->conteudo.letraFreq.frequencia =
         no1->conteudo.letraFreq.frequencia +
         no2->conteudo.letraFreq.frequencia;

         /* O filho esquerdo do novo n  o primeiro n */
         /* removido da lista e o filho direito  o     */
         /* segundo n removido da lista                */
      noNovo->conteudo.filhoEsquerda = &no1->conteudo;
      noNovo->conteudo.filhoDireita = &no2->conteudo;

         /* Insere o n na lista */
      InsereNoEmOrdem(aLista, noNovo);
   }

   return NULL; /* S para satisfazer o compilador... */
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria; i.e.,
 *                     libera o espao ocupado por cada
 *                     n da rvore
 *
 * Parmetros:
 *      arvore (entrada e sada) - ponteiro para a raiz
 *                                 da rvore binria a
 *                                 ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBin(tArvore arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBin(arvore->filhoEsquerda);

         /* Destri subrvore direita */
      DestroiArvoreBin(arvore->filhoDireita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * DestroiListaDeCodigos(): Libera o espao ocupado pelos
 *                          ns da lista de cdigos,
 *                          tornando-a vazia
 *
 * Parmetros:
 *      lista (entrada e sada) - endereo do ponteiro para
 *                                a lista que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiListaDeCodigos(tListaDeCodigos *lista)
{
   tListaDeCodigos p; /* Aponta para o prximo */
                      /* n a ser liberado     */

   if (!*lista) { /* Verifica se a lista est vazia */
      return; /* Lista vazia no precisa ser destruda */
   }

   p = *lista; /* Faz p apontar para o incio da lista */


      /* Visita cada n da lista liberando-o */
   do {
         /* Libera o espao ocupado pelo */
         /* array que armazena o cdigo  */
      free(p->codigo);

         /* Passa para o prximo n antes que o n corrente */
         /* seja destrudo. No importa que o ponteiro para */
         /*o incio da lista seja alterado. Afinal, este    */
         /* ponteiro ser mesmo anulado.                    */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

         /* Faz p apontar para o prximo n, cujo */
         /* endereo est armazenado em *lista    */
      p = *lista;
   } while (p);

      /* Neste ponto, a lista ficou vazia, pois o  */
      /* ltimo valor assumido por *lista foi NULL */
}
/****
 *
 * InsereEmListaDeCodigos(): Insere um novo n no incio da
 *                           lista encadeada contendo os
 *                           cdigos das letras
 *
 * Parmetros:
 *   *lista (entrada e sada) - ponteiro para a lista na
 *                               qual ser feita a insero
 *    letra (entrada) - letra a ser armazenada num novo n
 *    codigo (entrada) - array contendo o cdigo da letra
 *    n (entrada) - nmero de caracteres no cdigo da letra
 *
 * Retorno: Nada
 *
 ****/
void InsereEmListaDeCodigos( tListaDeCodigos *lista, int letra,
                             char *codigo, int n )
{
   tNoCodigo *ptrNovoNo; /* Apontar para o novo n alocado */
   int        i;

      /* Tenta alocar um novo n; se no for */
      /* possvel, aborta o programa         */
   ASSEGURA( ptrNovoNo = malloc(sizeof(tNoCodigo)),
             "Nao foi possivel alocar no" );

      /* Tenta alocar um array para armazenar o cdigo    */
      /* da letra; se no for possvel, aborta o programa */
   ASSEGURA( ptrNovoNo->codigo = malloc(n + 1),
             "Nao foi possivel alocar espaco para codigo" );

      /* Armazena a codificao da letra */
   for (i = 0; i < n; ++i)
      ptrNovoNo->codigo[i] = codigo[i];

      /* Transforma o cdigo em string */
   ptrNovoNo->codigo[i] = '\0';

      /* Armazena a letra no novo n */
   ptrNovoNo->letra = letra;

      /* O novo n apontar para o incio corrente da lista */
   ptrNovoNo->proximo = *lista;

      /* O incio da lista passa a apontar para o novo n */
   *lista = ptrNovoNo;
}

/****
 *
 * ColetaCodigos(): Coleta numa lista encadeada a codificao
 *                  de Huffman armazenada numa rvore binria
 *
 * Parmetros:
 *      raiz (entrada) - ponteiro para a raiz rvore contendo
 *                       a codificao
 *      codigo[] (entrada e sada) - array que armazena os
 *                                   cdigos dos caracteres
 *      n (entrada) - tamanho do array codigo[]
 *      listaDeCodigos (entrada e sada) - lista que armazenar
 *                                         as letras e seus
 *                                         respectivos cdigos
 *
 * Retorno: Nada
 *
 * Observao: Se voc no entendeu ainda, o caractere '0'
 *             deve ser entendido como o bit 0 e o caractere
 *             '1' deve ser entendido como o bit 1
 *
 ****/
void ColetaCodigos( tArvore raiz, char codigo[], int n,
                    tListaDeCodigos *listaDeCodigos )
{
      /* Armazena '0' no array quando encontra um filho   */
      /* esquerdo e chama a funo recursivamente para    */
      /* efetuar caminhamento na subrvore esquerda.      */
      /* Na prtica, usa-se o bit 0 e no o caractere '0' */
   if (raiz->filhoEsquerda) {
      codigo[n] = '0';
      ColetaCodigos( raiz->filhoEsquerda, codigo, n + 1,
                     listaDeCodigos );
   }
      /* Armazena '1' no array quando encontra um filho   */
      /* direito e chama a funo recursivamente para     */
      /* efetuar caminhamento na subrvore direita        */
      /* Na prtica, usa-se o bit 1 e no o caractere '1' */
   if (raiz->filhoDireita) {
      codigo[n] = '1';
      ColetaCodigos( raiz->filhoDireita, codigo, n + 1,
                     listaDeCodigos );
   }

      /* Se o n corrente for uma folha, */
      /* ela armazena uma letra...       */
   if (!(raiz->filhoEsquerda) && !(raiz->filhoDireita)) {
         /* Insere a letra armazenada nesta folha e seu */
         /* respectivo cdigo numa lista encadeada      */
      InsereEmListaDeCodigos( listaDeCodigos,
                              raiz->letraFreq.letra,
                              codigo, n);
   }
}

/****
 *
 * ExibeCodigos(): Exibe na tela cada letra e seu respectivo
 *                 cdigo armazenados numa lista encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para o primeiro n da lista
 *
 * Retorno: Nada
 *
 ****/
void ExibeCodigos(tListaDeCodigos lista)
{
      /* Se a lista est vazia, por   */
      /* que esta funo foi chamada? */
   ASSEGURA(lista, "A lista esta' vazia");

      /* Cabealho de apresentao na tela */
   printf( "\nCodigos"
           "\n=======\n");

      /* Visita cada n da lista exibindo */
      /* a informao contida no n       */
   while (lista) {
      printf("\n%c: %s", lista->letra, lista->codigo);

      lista = lista->proximo; /* Passa para o prximo n */
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * ExibeCodificacao(): Exibe na tela a codificao de um string
 *
 * Parmetros:
 *    string (entrada) - o string cuja codificao ser exibida
 *    listaDeCodigos (entrada) - lista encadeada contendo os
 *                               cdigos dos caracteres que
 *                               compem o string
 *
 * Retorno: Nada
 *
 ****/
void ExibeCodificacao( const char *string,
                       tListaDeCodigos listaDeCodigos )
{
   tListaDeCodigos p;

      /* Cabealho de apresentao na tela */
   printf( "\nCodificacao"
           "\n===========\n\n");

   while (*string) {
      p = listaDeCodigos;

      while (p && p->letra != *string) {
         p = p->proximo;
      }

         /* Se o lao terminou porque p  NULL, a letra no */
         /* foi encontrada. Portanto, deve haver algum erro */
         /* na lista de cdigos. Ento, aborta o programa.  */
      ASSEGURA(p, "Codigo de letra nao foi encontrado");

         /* p aponta para o n que contm */
         /* o cdigo da letra corrente    */
      printf("%s ", p->codigo);

      ++string; /* Passa para a prxima letra */
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * ExibeDecodificacao(): Exibe na tela a codificao de um string
 *
 * Parmetros:
 *    string (entrada) - o string cuja codificao ser exibida
 *    listaDeCodigos (entrada) - lista encadeada contendo os
 *                               cdigos dos caracteres que
 *                               compem o string
 *
 * Retorno: Nada
 *
 ****/
int EhFolha(tArvore arvore)
{
   return !arvore->filhoEsquerda && !arvore->filhoDireita;
}

/****
 *
 * ExibeDecodificacao(): Exibe na tela a decodificao de
 *                       um string
 *
 * Parmetros:
 *    string (entrada) - o string cuja decodificao ser
 *                       exibida
 *    arvore (entrada) - arvore contendo a codificao
 *
 * Retorno: Nada
 *
 ****/
void ExibeDecodificacao(const char *string, tArvore arvore)
{
   tArvore p;

      /* Cabealho de apresentao na tela */
   printf( "\nDecodificacao"
           "\n=============\n\n");

   while (*string) {
      p = arvore;

      do {
         if (*string == '0') {
            p = p->filhoEsquerda;
         } else if (*string == '1') {
            p = p->filhoDireita;
         } else {
            printf( "\nCaractere espurio "
                    "encontrado: %c\n", *string );
            exit(1);
         }

         ++string;
      } while (!EhFolha(p));

      putchar(p->letraFreq.letra);
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****************** Funes Acessrias ********************/

/****
 *
 * FrequenciasDeLetras(): Determina a frequncia de uma
 *                        letra do alfabeto da lingua
 *                        portuguesa
 *
 * Parmetros:
 *      aLetra (entrada) - inteiro que se supe estar
 *                         associado a uma letra cuja
 *                         frequncia se deseja determinar
 *
 * Retorno: A frequncia do caractere associado ao parmetro,
 *          se ele corresponter a uma letra; caso contrrio,
 *          um valor negativo.
 *
 * Observao: A frequncia de caracteres  baseada na pgina:
 *             https://pt.wikipedia.org/wiki/Alfabeto_portugues.
 *             A frequncia original no faz diferena entre
 *             letras maisculas e minsculas, o que  um
 *             destino quando se considera linguagem escrita.
 *             Portanto, aqui estimou-se (quer dizer,
 *             chutou-se mesmo) que a frequncia de uma letra
 *             maiscula  10% da frequncia da letra minscula
 *             correspondente. O interesse aqui  programao
 *             e no lingustica. Portanto, faa de conta que
 *             essa suposio  verdadeira. As frequncias do
 *             espao em branco e do hfen tambm foram
 *             totalmente chutadas. Se voc obtiver dados
 *             confiveis que possam substituir esses chutes,
 *             as substituies necessrias so fceis de fazer.
 *
 ****/
double FrequenciasDeLetras(int car)
{
   static tLetraFreq letraFreq[] = { {'a', 14.63}, {'b', 1.04},
                                     {'c', 3.88}, {'d', 5.01},
                                     {'e', 12.57}, {'f', 1.02},
                                     {'g', 1.30}, {'h', 1.28},
                                     {'i', 6.18}, {'j', 0.40},
                                     {'k', 0.02}, {'l', 2.78},
                                     {'m', 4.74}, {'n', 5.05},
                                     {'o', 10.73}, {'p', 2.52},
                                     {'q', 1.20}, {'r', 6.53},
                                     {'s', 7.81}, {'t', 4.34},
                                     {'u', 4.63}, {'v', 1.67},
                                     {'w', 0.01}, {'x', 0.21},
                                     {'y', 0.01}, {'z', 0.47},
                                     {'A', 14.63}, {'B', 1.04},
                                     {'C', 3.88}, {'D', 5.01},
                                     {'E', 12.57}, {'F', 1.02},
                                     {'G', 1.30}, {'H', 1.28},
                                     {'I', 6.18}, {'J', 0.40},
                                     {'K', 0.02}, {'L', 2.78},
                                     {'M', 4.74}, {'N', 5.05},
                                     {'O', 10.73}, {'P', 2.52},
                                     {'Q', 1.20}, {'R', 6.53},
                                     {'S', 7.81}, {'T', 4.34},
                                     {'U', 4.63}, {'V', 1.67},
                                     {'W', 0.01}, {'X', 0.21},
                                     {'Y', 0.01}, {'Z', 0.47},
                                     {' ', 5.00}, {'-', 0.50} }; /* Chute */
   int nCar = sizeof(letraFreq)/sizeof(letraFreq[0]);

   for (int i = 0; i < nCar; ++i) {
      if (letraFreq[i].letra == car) {
         return isupper(car) ? letraFreq[i].frequencia/10
                             : letraFreq[i].frequencia;
      }
   }

   return -1.0;
}

/****
 *
 * RemoveBrancos(): Remove espaos em branco de um string
 *
 * Parmetros:
 *      str (entrada/sada) - string a ser modificado
 *
 * Retorno: O endereo do string modificado
 *
 ****/
char *RemoveBrancos(char *str)
{
   char *p, /* Ponteiro auxiliar */
        *inicio; /* Guardar o incio do string */
      /* Faz p e 'inicio' apontarem para incio do string */
   p = inicio = str;
      /* Enquanto o caractere terminal no for       */
      /* encontrado, copia para o endereo apontado  */
      /* por p qualquer caractere que no seja igual */
      /* ao caractere que ser removido              */
   while (*str) {
         /* Se o caractere apontado por str no for   */
         /* ' ' ou '\t', copia-o para o endereo para */
         /* onde p aponta e faz p apontar para um     */
         /* caractere adiante                         */
      if (*str != ' ' && *str != '\t') {
         *p++ = *str;
      }
         /* Faz str apontar para o prximo caractere */
      str++;
   }
      /* Neste ponto, devido ao uso do incremento sufixo */
      /* na instruo if acima, p est apontando para um */
      /* caractere adiante do ltimo caractere copiado.  */
      /* Nesse local, o string deve ser terminado,       */
      /* copiando-se o caractere '\0'.                   */
   *p = '\0';
      /* O ponteiro 'inicio' guarda o endereo */
      /* do string recebido como parmetro     */
   return inicio;
}

/****
 *
 * ApresentaMenu(): Apresenta o menu de opes do programa
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 ****/
void ApresentaMenu(void)
{
   printf("\n\tOpcoes"
          "\n\t======\n"
          "\n1. Codifica caracteres"
          "\n2. Decodifica caracteres"
          "\n3. Exibe codigos"
          "\n4. Encerra programa"
          );
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa as funes da Codificao de Huffman
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 ****/
int main(void)
{
   char   *string; /* String digitado pelo usurio */
   tLista  listaNos; /* Apontar para a lista que armazena    */
                     /* temporariamente os ns da codificao */
   tArvore arvore;   /* A rvore que representa a codificao */
   char   *ar; /* Apontar para um array que     */
               /* armazena o cdigo de uma letra */
   int     op, /* Opo escolhida pelo usurio */
           nCaracteres, /* Nmero de caracteres armazenados */
                        /* no array ar apontado por ar      */
           nFolhas; /* Nmero de folhas da rvore */
                    /* de codificao             */

   tListaDeCodigos listaCodigos; /* Apontar para uma lista */
                                 /* encadeada contendo as   */
                                 /* letras e seus cdigos   */

      /* Apresenta o programa */
   printf( "\n\t>>> Este programa codifica a sequencia"
           "\n\t>>> de letras que voce introduzir.\n" );

      /* Cria a lista que armazena temporariamente */
      /*  os ns da codificao                    */
   listaNos = CriaListaEncadeadaOrd();

      /* O  nmero de folhas da rvore de codificao */
      /*  igual ao comprimento inicial da lista      */
   nFolhas = ComprimentoLE(listaNos);

      /* Cria a codificao do cara */
   arvore = Codifica(&listaNos);

      /*************************************************/
      /*                                               */
      /* A funo que exibe a codificao usa um array */
      /* para armazenar os cdigos  medida que visita */
      /* os ns da rvore de codificao. Portanto, o  */
      /* tamanho desse array deve ser pelo menos igual */
      /*  profundidade da rvore. Como toda rvore de */
      /* codificao de Huffman  estritamente binria */
      /* essa profundidade deve ser pelo menos igual   */
      /* ao nmero de folhas da rvore.                */
      /*                                               */
      /*************************************************/

   ar = malloc(nFolhas);

      /* Inicialmente a lista contendo os */
      /* cdigos das letras est vazia    */
   listaCodigos = NULL; /* Must! */

      /* Inicialmente, no h nenhum */
      /* caractere no array ar[]     */
   nCaracteres = 0; /* Must too! */

      /* Armazena os cdigos das letras numa lista */
   ColetaCodigos( arvore, ar, nCaracteres,
                  &listaCodigos );

   do { /* Lao principal do programa */
      ApresentaMenu();

      printf("\n\n\t>>> Escolha sua opcao: ");
      op = LeOpcao("1234");

      if (op == '4') {
         break;
      }

      switch (op) {
         case '1': /* Codifica um string */
               /* L o string */
            printf("\n\t>>> Digite uma sequencia de letras: ");
            string = LeLinhaIlimitada(NULL, stdin);

               /* Verifica se o usurio digitou apenas [ENTER] */
            if (!*string) {
               printf( "\n>>> Voce nao digitou nenhuma letra\n" );
               return 1; /* Este programa  intransigente */
            }

               /* Exibe na tela a codificao do string */
            ExibeCodificacao(string, listaCodigos);

               /* O espao ocupado pelo string lido foi alocado */
               /* dinamicamente e no  mais necessrio         */
            free(string);
         break;
         case '2': /* Decodifica um string */
               /* L o string */
            printf("\n\t>>> Digite o codigo: ");
            string = LeLinhaIlimitada(NULL, stdin);

               /* Verifica se o usurio digitou apenas [ENTER] */
            if (!*string) {
               printf( "\n>>> Voce nao digitou nada\n" );
               return 1; /* Este programa  intransigente */
            }

               /* Remove eventuais espaos em branco do string */
            RemoveBrancos(string);

               /* Exibe na tela a decodificao do string */
            ExibeDecodificacao(string, arvore);

               /* O espao ocupado pelo string lido foi alocado */
               /* dinamicamente e no  mais necessrio         */
            free(string);
            break;
         case '3': /* Exibe codigos */
               /* Exibe na tela os cdigos de letras obtidos */
            ExibeCodigos(listaCodigos);
            break;
         default:
            break;
      }
   } while (1);

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> Este programa codifica a sequencia
        >>> de letras que voce introduzir.

        Opcoes
        ======

1. Codifica caracteres
2. Decodifica caracteres
3. Exibe codigos
4. Encerra programa

        >>> Escolha sua opcao: 1

        >>> Digite uma sequencia de letras: banana

Codificacao
===========

1100010 100 11111 100 11111 100

        Opcoes
        ======

1. Codifica caracteres
2. Decodifica caracteres
3. Exibe codigos
4. Encerra programa

        >>> Escolha sua opcao: 2

        >>> Digite o codigo: 1100010 100 11111 100 11111 100

Decodificacao
=============

banana

        Opcoes
        ======

1. Codifica caracteres
2. Decodifica caracteres
3. Exibe codigos
4. Encerra programa

        >>> Escolha sua opcao: 3

Codigos
=======

n: 11111
d: 11110
 : 11101
p: 111001
E: 1110001
q: 1110000
m: 11011
u: 11010
t: 11001
O: 1100011
b: 1100010
f: 1100001
G: 1100000111
H: 1100000110
P: 110000010
N: 11000000
s: 1011
c: 10101
D: 10100111
-: 10100110
M: 10100101
U: 10100100
v: 101000
a: 100
r: 0111
i: 0110
e: 010
Q: 0011111111
J: 001111111011
w: 00111111101011
Y: 0011111110101011
W: 0011111110101010
K: 001111111010100
y: 0011111110100
X: 001111111001
k: 001111111000
x: 001111110
T: 00111110
j: 00111101
C: 00111100
A: 001110
l: 00110
S: 0010111
R: 0010110
g: 001010
h: 001001
B: 0010001111
F: 0010001110
V: 001000110
L: 00100010
I: 0010000
o: 000

        Opcoes
        ======

1. Codifica caracteres
2. Decodifica caracteres
3. Exibe codigos
4. Encerra programa

        >>> Escolha sua opcao: 4

**********************************************************/
